//----------------------------------------------------------------------------
// Random object spawner
// Christopher Rasmussen
// (c) University of Delaware, 2012
//----------------------------------------------------------------------------

#include "ros/ros.h"
#include "std_msgs/String.h"
#include "sensor_msgs/LaserScan.h"
#include "visualization_msgs/Marker.h"
#include "geometry_msgs/Twist.h"
#include "move_base_msgs/MoveBaseAction.h"
#include "nav_msgs/Odometry.h"
#include "nav_msgs/Path.h"

#include <gazebo_msgs/SetModelState.h>
#include <gazebo_msgs/SpawnModel.h>
#include <gazebo_msgs/DeleteModel.h>

#include <LinearMath/btQuaternion.h> // Needed to convert rotation ...
#include <LinearMath/btMatrix3x3.h>  // ... quaternion into Euler angles

#include <pcl/point_types.h>

#include <tf/transform_listener.h>

#include <iostream>
#include <sstream>

//----------------------------------------------------------------------------

using namespace std;

class Object_Parameters {
  
public:
  
  Object_Parameters(string s, 
		    double sx = 1.0, double sy = 1.0, double sz = 1.0, 
		    double x = 0.0, double y = 0.0, double z = 0.0, 
		    double roll = 0.0, double pitch = 0.0, double yaw = 0.0) 
  {
    collada_filename = s;
    scale_x = sx;  scale_y = sy;   scale_z = sz;
    pos_x = x;     pos_y = y;      pos_z = z;
    rpy_r = roll;  rpy_p = pitch;  rpy_y = yaw;
    times_spawned = 0;
  }
  
  string collada_filename;
  double scale_x, scale_y, scale_z;
  double pos_x, pos_y, pos_z;
  double rpy_r, rpy_p, rpy_y;
  int times_spawned;
  
};

//----------------------------------------------------------------------------

ros::NodeHandle *nhp;

ros::ServiceClient spawn_model_client;
gazebo_msgs::SpawnModelRequest spawnmodelrequest;
gazebo_msgs::SpawnModel spawnmodel;

vector <string> gazebo_color_array;

map <string, Object_Parameters *> large_object_params_map;
map <string, Object_Parameters *> small_object_params_map;
map <string, Object_Parameters *> obstacle_object_params_map;

//----------------------------------------------------------------------------

// See http://ros.org/wiki/simulator_gazebo/Tutorials/ListOfMaterials for 
// full list of colors

// if the same .dae file is to be used more than once, the obj_name must be different

// if dae_package_filename = "sphere", just uses sphere shortcut rather than loading a file
// same for "box", "cylinder"

string define_collada_model(string obj_name,
			    string dae_package_filename,
			    string gazebo_color,
			    double sx, double sy, double sz, 
			    double x, double y, double z, 
			    double rpy_r, double rpy_p, double rpy_y)
{
  stringstream ss, ss_origin, ss_geometry;
  
  ss << "<robot name=\"";
  ss << obj_name;
  ss << "\"> <link name=\"my_";
  ss << obj_name;
  ss << "\"> <inertial> <origin xyz=\"0 0 0\" /> <mass value=\"100.0\" /> <inertia  ixx=\"10\" ixy=\"0.0\"  ixz=\"0.0\"  iyy=\"100.0\"  iyz=\"0.0\"  izz=\"10\" /> </inertial> " << "<visual> ";
  ss_origin << "<origin xyz=\"" << x << " " << y << " " << z;

  if (dae_package_filename != "sphere" && dae_package_filename != "box" && dae_package_filename != "cylinder")
    ss_origin << "\" rpy=\"" << rpy_r << " " << rpy_p << " " << rpy_y;

  ss_origin << "\" /> ";
  ss_geometry << "<geometry> ";

  if (dae_package_filename == "sphere")
    ss_geometry << "<sphere radius=\"" << sx << "\" /> ";
  else if (dae_package_filename == "box")
    ss_geometry << "<box size=\"" << sx << " " << sy << " " << sz << "\" /> ";
  else if (dae_package_filename == "cylinder")
    ss_geometry << "<cylinder radius=\"" << sx << "\" length=\"" << sy << "\"/> ";
  else
    ss_geometry << "<mesh filename=\"package://" << dae_package_filename << "\" scale = \"" << sx << " " << sy << " " << sz << "\" /> ";

  ss_geometry << "</geometry> ";
  ss << ss_origin.str() << ss_geometry.str();
  ss << "</visual> <collision> ";
  ss << ss_origin.str() << ss_geometry.str();
  ss << "</collision> </link> ";
  ss << "<gazebo reference=\"my_";
  ss << obj_name;
  ss << "\"> <material>Gazebo/" << gazebo_color << "</material> ";
  ss << "<static>true</static> ";
  ss << "<turnGravityOff>true</turnGravityOff> ";
  ss << "</gazebo> </robot>"; 

  //cout << ss.str() << endl;

  return ss.str();
}

//----------------------------------------------------------------------------

// pick a random object from the map structure supplied, give it a random color
// and yaw, and spawn it at (x, y, z).  for some reason the chair object doesn't like
// getting spun around, so we don't change its yaw

void spawn_random_object(double x, double y, double z, map <string, Object_Parameters *> & object_params_map)
{
  int index, rand_index;
  map <string, Object_Parameters *>::iterator iter;
  stringstream ss;
  string name, name_prefix;
  Object_Parameters *params;
  double theta;

  spawn_model_client = nhp->serviceClient<gazebo_msgs::SpawnModelRequest>("/gazebo/spawn_urdf_model");

  spawnmodel.request.robot_namespace = "gazebo_worlds";
//   spawnmodel.request.initial_pose.orientation.x = 0;
//   spawnmodel.request.initial_pose.orientation.y = 0;
//   spawnmodel.request.initial_pose.orientation.z = 0;
//   spawnmodel.request.initial_pose.orientation.w = 1;
  spawnmodel.request.reference_frame = "map";

  rand_index = rand() % object_params_map.size();
  
  for (iter = object_params_map.begin(), index = 0; index < rand_index; iter++, index++)
    ;
  
  name_prefix = iter->first;
  params = iter->second;
  
  ss.str("");
  ss << name_prefix << "_" << params->times_spawned++;
  name = ss.str();
  
  spawnmodel.request.model_name = ss.str();
  spawnmodel.request.model_xml = define_collada_model(name,
						      params->collada_filename,
						      gazebo_color_array[rand() % gazebo_color_array.size()],
						      params->scale_x, params->scale_y, params->scale_z, 
						      params->pos_x, params->pos_y, params->pos_z,
						      params->rpy_r, params->rpy_p, params->rpy_y);
  
  spawnmodel.request.initial_pose.position.x = x;
  spawnmodel.request.initial_pose.position.y = y;
  if (name_prefix == "sphere")
    spawnmodel.request.initial_pose.position.z = z + params->scale_x;
  else if (name_prefix == "box")
    spawnmodel.request.initial_pose.position.z = z + 0.5 * params->scale_x;
  else if (name_prefix == "cylinder")
    spawnmodel.request.initial_pose.position.z = z + 0.5 * params->scale_y;
  else
    spawnmodel.request.initial_pose.position.z = z;

  // random yaw rotation

  if (name_prefix != "chair")
    theta = fmod(rand(), M_PI);
  else
    theta = 0.0;
    
  btQuaternion q(theta, 0, 0);
  spawnmodel.request.initial_pose.orientation.x = q.x();
  spawnmodel.request.initial_pose.orientation.y = q.y();
  spawnmodel.request.initial_pose.orientation.z = q.z();
  spawnmodel.request.initial_pose.orientation.w = q.w();

  if (spawn_model_client.call(spawnmodel)) {
    printf("(%.0lf, %.0lf): %s SUCCESS\n", x, y, name.c_str());

    // if that was a table or stepfield then put something on it?

    /*
    if (name_prefix == "table")
      spawn_random_object(x, y, 0.28, small_object_params_map);
    else if (name_prefix == "stepfield")
      spawn_random_object(x, y, 0.08, small_object_params_map);
    */
  }
  else 
    printf("(%.0lf, %.0lf): %s FAIL\n", x, y, name.c_str());
  
}

//----------------------------------------------------------------------------

void spawn_everything()
{
  double x, y;

  // this grid arrangment is customized to the hector indoor environment

  for (x = -3; x <= 3; x += 6)
    for (y = 0; y <= 6; y += 3)
	spawn_random_object(x, y, 0.01, small_object_params_map);

  spawn_random_object(0, 3, 0.01, obstacle_object_params_map);

  exit(1);
}

//----------------------------------------------------------------------------

void initialize_models()
{
  gazebo_color_array.push_back("Red");
  gazebo_color_array.push_back("Green");
  gazebo_color_array.push_back("Yellow");
  gazebo_color_array.push_back("Blue");

  //  large_object_params_map["chair"] = new Object_Parameters("spawner/meshes/chair.dae", 0.0012, 0.0012, 0.0012, -0.35, -0.25, 0);
  
  small_object_params_map["sphere"] = new Object_Parameters("sphere", 0.25, 0.25, 0.25);
  small_object_params_map["box"] = new Object_Parameters("box", 0.4, 0.4, 0.4);
  small_object_params_map["cylinder"] = new Object_Parameters("cylinder", 0.25, 0.6, 0.5);
  small_object_params_map["rock"] = new Object_Parameters("spawner/meshes/rock_2.dae", 0.001, 0.002, 0.0025);
  small_object_params_map["duck"] = new Object_Parameters("spawner/meshes/duck.dae", 0.003, 0.003, 0.003, 0, 0, 0, 1.5708, 0, 0);

  obstacle_object_params_map["table"] = new Object_Parameters("spawner/meshes/table.dae", 0.7, 0.7, 0.2);
  obstacle_object_params_map["stepfield"] = new Object_Parameters("spawner/meshes/stepfield.dae", 0.8, 0.8, 0.1);
  obstacle_object_params_map["ramp"] = new Object_Parameters("spawner/meshes/reduced_full_ramp.dae", 0.4, 0.4, 0.5);

  // rviz runs a bit slow with these in the environment

  /*
  small_object_params_map["football"] = new Object_Parameters("spawner/meshes/football.dae", 0.04, 0.04, 0.04, 0, 0, -0.5);
  small_object_params_map["soccerball"] = new Object_Parameters("spawner/meshes/soccerball.dae", 0.002, 0.002, 0.002);
  small_object_params_map["teddy"] = new Object_Parameters("spawner/meshes/teddy.dae", 0.015, 0.015, 0.015);
  */
}

//----------------------------------------------------------------------------

int main(int argc, char **argv)
{
  int i;

  ros::init(argc, argv, "objects");

  srand (time(NULL));

  nhp = new ros::NodeHandle();

  initialize_models();
  spawn_everything();

  ros::Rate loop_rate(10);

  while (ros::ok()) {

    ros::spinOnce();
    loop_rate.sleep();

  }

  return 0;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
